from nio.router.base import BlockRouter
from nio.util.runner import RunnerStatus
from nio.util.threading import spawn


class ThreadedBlockRouter(BlockRouter):

    """A router that delivers signals asynchronously"""

    def deliver_signals(self, block_receiver, signals):
        """ Delivers signals to given block in a new thread """
        spawn(self._safe_process_signals, block_receiver, signals)

    def _safe_process_signals(self, block_receiver, signals):
        try:
            self.notify_signals_to_block(block_receiver, signals)
        except:
            self.status.add(RunnerStatus.error)
            self.logger.exception(
                "{0}.process_signals failed".format(block_receiver.block.name))



class Request(object):

    """ Encapsulates web request information

    Create a Request instance to include web request parameters

    """

    def __init__(self, body, params, headers):
        """ Create a new Request instance.

        Args:
            body: request body
            params: request params
            headers: request headers

        """
        self._identifier = params.get('identifier')
        self._body = body
        self._params = params

        self._headers = headers

    def get_body(self):
        """ Retrieves request body

        Returns:
            Request body

        """
        return self._body

    def get_identifier(self):
        """ Retrieves request identifier

        Returns:
            Request identifier

        """
        return self._identifier

    def get_params(self):
        """ Retrieves request params

        Returns:
            Request params

        """
        return self._params

    def get_header(self, header, default=None):
        """ Retrieves request headers

        Returns:
            Request headers

        """
        return self._headers.get(header, default)



from nio.modules.web import WebEngine
from nio.modules.module import Module


class WebModule(Module):

    def proxy_web_class(self, web_class):
        WebEngine.proxy(web_class)

    def finalize(self):
        WebEngine.unproxy()
        super().finalize()

    def get_module_order(self):
        return 40



from nio.types.base import Type


class BoolType(Type):

    @staticmethod
    def serialize(value, **kwargs):
        """ Convert a value to a JSON serializable value """
        return value

    @staticmethod
    def deserialize(value, **kwargs):
        """ Convert value to bool"""
        try:
            return bool(value)
        except:
            raise TypeError("Unable to cast value to bool: {}".format(value))




class WebEngine():

    @classmethod
    def get(cls, port, host, config=None):
        pass

    @classmethod
    def remove_server(cls, server):
        pass

    @classmethod
    def get_servers(cls):
        pass

    @classmethod
    def configure(cls, configuration):
        """ Configure Web Server """
        pass

    @classmethod
    def get_name(cls):
        """ Returns server name """
        pass

    @classmethod
    def get_version(cls):
        """ Returns web server version """
        pass

    @classmethod
    def start(cls, callback=None):
        """ Starts Web Server using configured params """
        pass

    @classmethod
    def stop(cls):
        """ Stops running Web Servers """
        pass

    @classmethod
    def block(cls):
        """ Blocks the Web Server until further request arrives """
        pass



from nio.modules.context import ModuleContext
from nio.util.scheduler.job import Job
from nio.util.scheduler.scheduler import Scheduler
from nio.modules.scheduler.module import SchedulerModule


class TestingSchedulerModule(SchedulerModule):

    def initialize(self, context):
        super().initialize(context)
        self.proxy_job_class(Job)

        Scheduler.configure(context)
        Scheduler.start()

    def finalize(self):
        Scheduler.shutdown()
        super().finalize()

    def prepare_core_context(self):

        context = ModuleContext()
        context.min_interval = 0.1
        context.resolution = 0.1

        return context



from nio.modules.proxy import ModuleProxy


class Persistence(ModuleProxy):

    """ Persistence Module

    This class encapsulates the user-facing interface to NIO's
    persistence layer. Block writers use this module to save dynamic
    data at runtime, allowing it to persist in the block instance
    after a service/instance restart.

    Args:
        name (str): The name of the block whose data will be persisted.

    Example:
        self.persistence.store('foo', 'bar')
        self.persistence.save() # saves the stored k/v store to disk
        val = self.persistence.load('foo') # now val == 'bar'

    """

    # Note: We don't proxy a constructor, the ModuleProxy __init__ will handle
    # the mapping to the proper class for us, so just call super

    def __init__(self, name):
        """ Create a Persistence object for a given name """
        super().__init__(name)

    def store(self, key, value):
        """ Store a key/value datum in memory.

        Args:
            key: The key to store into
            value: The value associated with that key
        """
        raise NotImplementedError()

    def load(self, key, default=None):
        """ Load a value from the k/v store *currently in memory*.

        Args:
            key: The key to lookup in the store.
            default:  The value to return if the key is not present

        Returns:
            value: The value associated with that key

        """
        raise NotImplementedError()

    def has_key(self, key):
        """ Check whether a particular key exists in the persistence store

        Args:
            key: The key in question.

        Returns:
            exists (bool)

        """
        raise NotImplementedError()

    def clear(self, key):
        """ Remove the given key and associated value from the store

        As above, this will not be reflected permanently until
        Persistence.save is called.

        Args:
            key: The key pointing to the data to clear.
        """
        raise NotImplementedError()

    def save(self):
        """ Save the in-memory store to persistence.

        This allows the data therein to persist between instance/service
        restarts.
        """
        raise NotImplementedError()

    @classmethod
    def setup(cls, configuration):
        """ Class method to initialize the Persistence class with a config.

        This method will be called after the ModuleProxy is initialized.

        Args:
            configuration (dict): The dictionary that
                module_init was called with

        Returns:
            None
        """
        raise NotImplementedError()

    @classmethod
    def configure(cls, service_name):
        """ Class method to configure the Persistence class with a service

        This allows us to have a service-by-service namespace for
        persisted block data.

        Args:
            service_name (str): The name of the service constituting this
                                process.

        Returns:
            None

        """
        pass



""" The data that a block will be configured with """


class BlockContext(object):

    def __init__(self, block_router, properties,
                 service_name='', command_url=''):
        """ Initializes information needed for a Block

        This BlockContext will be passed to the `configure` method on each
        block instance before getting started.

        Args:
            block_router (BlockRouter): The router in which the
                block will be running. The only requirements are that it
                can handle signals notified by its blocks.

            properties (dict): The block properties (metadata) that
                will be deserialized and loaded as block properties.
            service_name (str): The name of the service this block belongs to
            command_url (str): The URL at which this block can be commanded.
                This URL will not have host or port information, as that may
                be different based on public/private IP. It will look like
                "/services/ServiceName/BlockAlias/"

        """

        self.block_router = block_router
        self.properties = properties
        self.service_name = service_name
        self.command_url = command_url



from nio.block.mixins.retry.strategy import BackoffStrategy


class Retry(object):

    """ A block mixin that provides retry functionality.

    By including this mixin, your block will have access to a method that can
    retry on failure. This is useful when performing tasks that have a chance
    of failing but can then work upon retrying. Example use cases are database
    queries or other requests over a network.

    The block developer should select a retry backoff strategy from the
    pre-built strategies in nio.block.mixins.retry.strategies or can define
    their own according to the BackoffStrategy spec

    How to use this mixin:
        1. In your block's configure method, define your backoff strategy using
        the use_backoff_strategy method.
        2. Call execute_with_retry with the function that you want to execute.
        If the target function raises an exception, it will retry until it
        either succeeds or the backoff strategy has decided it should stop
        retrying. If that occurs, execute_with_retry will raise the exception
        that the target function raised originally.
        3. Optionally, override before_retry to write custom code that will be
        performed before attempting a retry.
    """

    def __init__(self):
        super().__init__()
        # Use the placeholder backoff strategy in case the block developer
        # doesn't define one. Note, this strategy will never retry
        self._backoff_strategy = BackoffStrategy(logger=self.logger)

    def execute_with_retry(self, execute_method, *args, **kwargs):
        """ Execute a method and retry if it raises an exception

        Args:
            execute_method (callable): A function to attempt to execute. The
                function may be called multiple times if retries occur.
            args/kwargs: Optional arguments to pass to the execute method

        Returns:
            The result of execute_method upon success.

        Raises:
            Exception: The exception that execute_method raised when the
                backoff strategy decided to stop retrying.
        """
        # Save a stringified version of the method's name
        # If it doesn't define __name__, use however we should stringify
        execute_method_name = getattr(
            execute_method, '__name__', str(execute_method))
        while True:
            try:
                result = execute_method(*args, **kwargs)
                # If we got here, the request succeeded, let the backoff
                # strategy know then return the result
                self._backoff_strategy.request_succeeded()
                return result
            except Exception as exc:
                self.logger.warning(
                    "Retryable execution on method {} failed".format(
                        execute_method_name, exc_info=True))
                self._backoff_strategy.request_failed(exc)
                should_retry = self._backoff_strategy.next_retry()
                if not should_retry:
                    # Backoff strategy has said we're done retrying,
                    # so re-raise the exception
                    self.logger.exception(
                        "Out of retries for method {}.".format(
                            execute_method_name))
                    raise
                else:
                    # Backoff strategy has done its waiting and has instructed
                    # us to retry again. Execute any pre-work before looping
                    # and executing the method again
                    self.before_retry(*args, **kwargs)

    def use_backoff_strategy(self, strategy, *args, **kwargs):
        """ Tell this mixin which backoff strategy to use.

        This method should be called in a block's configure method.

        Args:
            strategy (class): A subclass of BackoffStrategy to use in this
                block when retrying
            args/kwargs: Optional arguments to pass to the constructor of the
                backoff strategy
        """
        if not issubclass(strategy, BackoffStrategy):
            raise TypeError("Backoff strategy must subclass BackoffStrategy")
        self._backoff_strategy = strategy(*args, **kwargs)
        self._backoff_strategy.use_logger(self.logger)

    def before_retry(self, *args, **kwargs):
        """ Perform any actions before the next retry occurs.

        Override this in your block to take action before a retry is attempted.
        This would be the function where you would make any reconnections or
        any other behavior that may make the next retry attempt succeed.
        """
        pass  # pragma: no cover



from datetime import datetime
import logging


class NIOFilter(logging.Filter):

    def filter(self, record):
        """  Filter the record before being logged

        We add niotime and context properties to the record so loggers
        created before the Logging Module initialization can be written
        using nio formatter
        """
        if not hasattr(record, 'niotime'):
            record.niotime = self._time()
            record.context = record.name
        return True

    def _time(self):
        """ Generate a nicely formatted time string for a log entry

        Args:
            None

        Returns:
            A time string
        """

        now = datetime.utcnow()
        return '%d-%02d-%02d %02d:%02d:%02d.%03d' % (
            now.year, now.month, now.day,
            now.hour, now.minute, now.second,
            now.microsecond / 1000)



class LooseMatching(object):

    @classmethod
    def matches(cls, sub_topics, pub_topics):
        """ Finds out if there is a match between sub and pub topics

        Args:
            sub_topics: subscriber topics
            pub_topics: publisher topics

        For it to match, there should be a key where the publisher
        publishes something the subscriber cares about
        """

        # empty lists, publish everything, subscribe everything
        if not len(sub_topics) and not len(pub_topics):
            return True

        matching_keys = [key for key in sub_topics.keys()
                         if key in pub_topics.keys()]
        if not matching_keys:
            return False

        for key, sub_values in sub_topics.items():
            if key not in pub_topics:
                continue

            # if publishing everything from key, subscriber is interested
            if not len(pub_topics[key]):
                return True

            # if subscriber list is empty, it means subscribes to
            # everything being published by key
            if not len(sub_values):
                return True

            match_for_key = False
            for value in pub_topics[key]:
                # publishing something subscriber cares about is enough
                # for a match
                if value in sub_values:
                    match_for_key = True
                    break

            # a common key existed, but not a single match
            if not match_for_key:
                return False

        return True



"""
  NIO block support base class

"""
from collections import defaultdict
from nio.testing.test_case import NIOTestCase
from nio.router.base import BlockRouter
from nio.block.context import BlockContext
from nio.block.terminals import DEFAULT_TERMINAL
from nio.signal.status import StatusSignal


class BlockRouterForTesting(BlockRouter):

    """ A testing router to be used in block test cases.

    Supports block tests by keeping track of signal counts per block and
    block status as provided by management signals.
    """

    def __init__(self):
        super().__init__()
        self._test_case = None
        self._block_signal_counts = {}
        self._block_mgmt_signal_counts = defaultdict(int)
        self._block_status = {}

    def configure(self, test_case):
        """ Configures router by saving the test case being supported

        Args:
            test_case (NIOTestCase): test case to call when notifying signals

        """
        self._test_case = test_case

    def notify_signals(self, block, signals, output_id):
        """ Receives block signals.

         Keeps track of the signal count per block and per output id
         Forwards notification to test_case's 'signals_notified' method

        Args:
            block (Block): notifier block
            signals (list): signals being notified
            output_id: output identifier

        """
        if output_id is None:
            output_id = DEFAULT_TERMINAL

        self._block_signal_counts[block] = \
            self._block_signal_counts.get(block) or defaultdict(int)
        self._block_signal_counts[block][output_id] += len(signals)

        self._test_case.signals_notified(block, signals, output_id)

    def notify_management_signal(self, block, signal):
        """ Receives management signal.

        Keeps track of the signal count per block
        Forwards signal to test_case's 'management_signal_notified' method

        Args:
            block (Block): notifier block
            signal (Signal): signal being notified

        """
        if isinstance(signal, StatusSignal):
            self._block_status[block] = signal.status

        self._block_mgmt_signal_counts[block] += 1
        self._test_case.management_signal_notified(block, signal)

    def get_signals_from_block(self, block=None, output_id=DEFAULT_TERMINAL):
        """ Provides signal count per block or for all blocks

        Args:
            block (Block): block instance
            output_id: output identifier

        Returns:
            The number of signals a block has notified.
            If block is None, return the total number of signals notified

        """
        if block is None:
            return sum([sum(blk.values()) for blk
                        in self._block_signal_counts.values()])

        block_counts = self._block_signal_counts.get(block)
        if block_counts is None:
            return 0

        return block_counts.get(output_id) or 0


class NIOBlockTestCase(NIOTestCase):

    """ Base Unit Test case class for blocks

    Allows block tests to see the signals that get notified and access some
    other data about them by setting up a router.

    Simply create the block you want to use and call configure_block with the
    properties the block should have.

    Example:
    b = MyTestableBlock()
    self.configure_block(b, {"loglevel": "DEBUG"})

    Now, whenever b notifies signals, the signals_notified function will be
    called in the test case.
    """

    def __init__(self, methodName='runTests'):
        super().__init__(methodName)
        self._router = BlockRouterForTesting()
        self.last_notified = defaultdict(list)

    def setUp(self):
        super().setUp()
        self._router.configure(self)

    def tearDown(self):
        super().tearDown()

    def configure_block(self, block, block_properties):
        """ Configures block by assigning properties and router to given block

        Args:
            block (Block): block instance
            block_properties (dict): properties to assign

        """
        # Blocks should always have a 'name', but we'll let it pass in tests
        block_properties["name"] = block_properties.get("name", "default")
        block.configure(BlockContext(
            self._router,
            block_properties,
            'TestSuite',
            ''))

    def signals_notified(self, block, signals, output_id):
        """ Receives block signals notification

        Args:
            block (Block): notifying block
            signals (list): signals being notified
            output_id: output identifier

        """
        self.last_notified[output_id].extend(signals)

    def management_signal_notified(self, block, signal):
        """ Receives block management signal notification

        Args:
            block (Block): notifying block
            signal (Signal): signal being notified
        """
        pass

    def assert_num_signals_notified(self, num, block=None,
                                    output_id=DEFAULT_TERMINAL):
        """ Assert that the number of signals notified is a certain number.

        Args:
            num (int): number to check against
            block (Block): the block which should have notified. If not
                included, all blocks will be considered.
            output_id: The output id of the block to consider.
        """
        signals = self._router.get_signals_from_block(block, output_id)
        self.assertEqual(num, signals)

    def assert_num_mgmt_signals_notified(self, num, block=None):
        """ Assert the number of management signals notified is a number.

        Args:
            num (int): number to check against
            block (Block): the block which should have notified. If not
                included, all blocks will be considered.
        """
        if block:
            self.assertEqual(
                num, self._router._block_mgmt_signal_counts[block])
        else:
            self.assertEqual(
                num, sum(self._router._block_mgmt_signal_counts.values()))

    def assert_block_status(self, block, status):
        """ Asserts that the given block has a certain block status

        Blocks will have a status of '' if they have not had a status
        reported yet.

        Args:
            block (Block): The block to check the status of
            status (BlockStatus.xxx): The BlockStatus enum value to have
        """
        self.assertEqual(self._router._block_status.get(block, ''), status)



class Response(object):

    """ Encapsulates web response information

    Create a Response instance to include web response information

    """

    def set_header(self, header_name, header_value):
        """ Sets a single HTTP header.

        Args:
            header_name: header name
            header_value: header value

        """
        raise NotImplementedError()

    def set_status(self, status, message=None):
        """ Sets a web response status.

        Args:
            status: http status
            message: optional message

        """
        raise NotImplementedError()

    def set_body(self, body):
        """ Sets a web response body.

        Args:
            body: response body

        """
        raise NotImplementedError()





class Persistence(object):

    """ An implementation of the Persistence proxy that works in memory.

    This will not actually persist any values, but will keep them in
    memory to make sure that save/load commands work properly.

    In other words, this will "persist" its data only as long as the current
    environment is active. This is useful in unit-tests and in environments
    where persistence is not desired, but is hardly useful in practice.
    """

    def __init__(self, name):
        self._name = name
        self._values = {}

    def store(self, key, value):
        self._values[key] = value

    def load(self, key, default=None):
        return self._values.get(key, default)

    def has_key(self, key):
        return key in self._values

    def clear(self, key):
        if key in self._values:
            del self._values[key]

    def save(self):
        pass

    @classmethod
    def setup(cls, configuration):
        pass

    @classmethod
    def configure(cls, service_name):
        pass



import heapq
from uuid import uuid4
from datetime import timedelta
from threading import Event, RLock

from nio.util.logging import get_nio_logger
from nio.util.threading import spawn
from time import time, sleep
from collections import namedtuple

QueueEvent = namedtuple('Event', 'time, id, target, frequency, args, kwargs')


class SchedulerHelper(object):

    def __init__(self, resolution, min_delta):
        """ Initializes scheduler.

        Args:
            resolution: Specifies the scheduler resolution, this value
                is used during sleeping operations when it is determined
                a wait is needed before determining the next event to be
                executed
            min_delta: minimum value to consider when scheduling a task
        """
        super().__init__()

        self._resolution = resolution
        self._min_delta = min_delta

        self.logger = get_nio_logger("Custom Scheduler")
        self._queue = []
        self._queue_lock = RLock()
        self._stop_event = Event()
        self._events = dict()
        self._events_lock = RLock()

    def start(self):
        """ Starts scheduler.

        """

        while not self._stop_event.is_set():

            try:
                event = None
                with self._queue_lock:
                    queue_length = len(self._queue)
                    self.logger.debug('Queue contains: {0} tasks'.
                                      format(queue_length))

                    if queue_length:
                        # have access to first in time to execute event
                        event = heapq.heappop(self._queue)

                if not event:
                    sleep(self._resolution)
                    continue

                # if there are events, check their time to see if
                # it is up for firing it.
                event_time, event_id, target, frequency, args, kwargs = event
                now = time()
                # find out if need to sleep or execute event
                if now < event_time:
                    self.logger.debug(
                        'Current task time has not been reached, {0} remains'.
                        format(event_time - now))

                    with self._queue_lock:
                        # push it back when event is not ready
                        heapq.heappush(self._queue, event)
                    # do not allow big sleeps, have some control over it
                    # in case scheduler is stopped
                    delay = min(event_time - now, self._resolution)
                    sleep(delay)
                else:
                    # time is up, execute
                    try:
                        self.logger.debug("Executing: {0}".format(target))
                        # launch target task from a different thread thus
                        # making scheduler independent from task duration
                        spawn(target, *args, **kwargs)
                    except Exception:
                        self.logger.exception('Calling: {0}'.format(target))

                    with self._events_lock:
                        # before processing any further, make sure event has
                        # not been cancelled
                        if event_id in self._events:
                            # is it repeatable?
                            if frequency:
                                # reschedule it back, adding frequency to
                                # event time
                                event = QueueEvent(event_time + frequency,
                                                   event_id,
                                                   target,
                                                   frequency,
                                                   args, kwargs)
                                # housekeeping new event in
                                with self._queue_lock:
                                    heapq.heappush(self._queue, event)
                                self._events[event_id] = event
                            else:
                                # remove event when not repeatable
                                self._events.pop(event_id)
                        else:
                            self.logger.debug("Event: {0} was cancelled".
                                              format(event_id))

            except Exception:
                # log any exception in this big try/except
                # and do not leave loop
                self.logger.exception('Exception caught')

    def stop(self):
        """ Stops scheduler.

        """
        self._stop_event.set()

    def add(self, target, delta, repeatable, *args, **kwargs):
        """ Adds an event to the scheduler.

        Args:
            target (callable): The task to be scheduled.
            delta (timedelta): The scheduling delta. The scheduler will
                wait for this delta before running 'target' for the first
                time.
            repeatable (bool): When False, 'target' is run only once.
                Otherwise, it is run repeatedly at an delta defined by
                'delta'.
            args: Positional arguments to be passed to 'target'.
            kwargs: Keyword arguments to be passed to 'target'.

        Returns:
            Event Identifier, which can be used for cancellation

        """
        if not isinstance(delta, timedelta):
            raise AttributeError('delta must be of type: timedelta')

        delta = delta.total_seconds()
        if repeatable:
            # make sure delta is not smaller than minimum
            if delta < self._min_delta:
                delta = self._min_delta
            frequency = delta
        else:
            # when non repeatable, allow delta to be as small as user wishes
            # it to be
            frequency = 0

        event_id = uuid4().hex
        event = QueueEvent(time() + delta, event_id, target, frequency,
                           args, kwargs)

        # add to queue
        with self._queue_lock:
            heapq.heappush(self._queue, event)

        # add to events
        with self._events_lock:
            self._events[event_id] = event

        return event_id

    def cancel(self, event_id):
        """ Cancels an event scheduled for execution.

        Args:
            event_id: Event Id obtained from an "add" call

        Returns:
            True: If event was previously scheduled
            False: otherwise

        """
        # remove it from events dictionary
        event = None
        with self._events_lock:
            if event_id in self._events:
                event = self._events.pop(event_id)
        if event:
            try:
                with self._queue_lock:
                    self._queue.remove(event)
                    heapq.heapify(self._queue)
                self.logger.debug('Success cancelling event')
                return True
            except Exception:
                self.logger.debug('Failure to remove event {0} from queue'
                                  ' while cancelling a job'.format(event))
        return False



class DefaultMatching(object):

    @classmethod
    def matches(cls, sub_topics, pub_topics):
        """ Finds out if there is a match between sub and pub topics

        Args:
            sub_topics: subscriber topics
            pub_topics: publisher topics

        For it to match, every key/value the subscriber specifies
        must have a matching in the publisher and no more

        If publisher provides more than what the subscriber is looking for
        there is no match.
        """
        for key, sub_values in sub_topics.items():
            # no match, if a subscriber key does not exist in the publisher
            if key not in pub_topics:
                return False

            # if subscriber does not specified values for given key,
            # assume that it is interested in everything published by key
            if not len(sub_values):
                continue

            # when values for a given key are specified by the subscriber,
            # then all the values provided by publisher need to be present
            # in the subscriber topics, in other words, publisher values for a
            # given key are a subset of subscriber values for same key
            for value in pub_topics[key]:
                if value not in sub_values:
                    return False

        return True



class BackoffStrategy(object):

    def __init__(self, logger=None):
        super().__init__()
        self.logger = logger
        self.retry_num = 0

    def request_failed(self, exc):
        """ Called when a retryable request fails """
        self.retry_num += 1

    def request_succeeded(self):
        """ Called when a retryable request succeeds """
        self.retry_num = 0

    def next_retry(self):
        """ Do any waiting then return True to execute the retry.

        This function is responsible for determining whether or not a retry
        should take place, and if so, how long (if at all) to wait before
        doing so. If the strategy chooses to wait, it should just call sleep
        inside of this method since it will likely be running in its own
        thread.

        Returns:
            bool: True if the retry should execute. False if we are "done"
                retrying
        """
        self.logger.warning("No retry strategy specified, not retrying")
        return False

    def use_logger(self, logger):
        """Use a logger instance in this backoff strategy.

        This helper function is useful for blocks or mixins to call so that the
        backoff strategy can log from that block. This gets called
        automatically by the Retry mixin but you are free to make another call
        to use a different logger.

        Args:
            logger (Logger): A logger that the backoff strategy can use
        """
        self.logger = logger



class WebServer(object):
    """ Web Server

    Serves HTTP requests given a host, port, and additional configuration

    """

    @property
    def host(self):
        """ Provides server host """
        return None

    @property
    def port(self):
        """ Provides server port """
        return None

    def add_handler(self, handler):
        """ Adds a web handler

        Args:
            handler (WebHandler): Web handler

        """
        raise NotImplementedError()

    def start(self, configuration):
        """ Starts web server

        Args:
            configuration: Server configuration

        """
        raise NotImplementedError()

    def stop(self):
        """ Stops web server

        """
        raise NotImplementedError()



from time import sleep
from nio.block.mixins.retry.strategy import BackoffStrategy


class LinearBackoff(BackoffStrategy):

    def __init__(self, max_retries=-1, multiplier=1):
        """ Create an instance of a linear backoff strategy

        Args:
            max_retries (int): The maximum number of retries to make before
                giving up on retrying.
            multiplier (float): The multiplier to apply when sleeping. By
                default, this strategy will sleep a number of seconds equal to
                the current retry number. This argument allows you to alter
                that. For example, by default the 5th retry will sleep for 5
                seconds. Passing a multiplier of 1.5 will cause the 5th retry
                to sleep for 7.5 seconds instead.
        """
        super().__init__()
        self._max_retries = max_retries
        self._multiplier = multiplier

    def next_retry(self):
        """Sleep a certain number of seconds unless we are out of retries."""
        # Check if we have a maximum number of retries and we've passed it
        self.logger.info("Executing retry number {}, max retries is {}".format(
            self.retry_num, self._max_retries))
        if self._max_retries >= 0 and self.retry_num > self._max_retries:
            self.logger.warning(
                "Retry number {} exceeds the  max retries of {}".format(
                    self.retry_num, self._max_retries))
            return False

        # We will wait the current retry number times the multiplier,
        # so first retry will be 1*mult, second will be 2*mult, etc
        sleep(self.retry_num * self._multiplier)
        return True



"""

    Common API Security

"""
import threading
from nio.modules.security.task import SecureTask
from nio.modules.security.authorizer import Authorizer, Unauthorized


class protected_access(object):

    """ Decorator to add to a method to protect access

    Retrieve user from current thread and evaluate access
    """

    def __init__(self, *tasks, meet_all=True):
        self._tasks = [task if isinstance(task, SecureTask)
                       else SecureTask(task)
                       for task in tasks]
        self._meet_all = meet_all

    def __call__(self, f):
        def wrapped_f(*args):
            # retrieve user
            thread = threading.current_thread()
            user = getattr(thread, "user", None)
            if user is None:
                raise Unauthorized()

            # evaluate all of the tasks
            Authorizer.authorize_multiple(
                user, *self._tasks, meet_all=self._meet_all)
            return f(*args)
        return wrapped_f


def _set_user(user):
    setattr(threading.current_thread(), "user", user)



from nio.properties import TimeDeltaProperty, BoolProperty
from nio.modules.persistence import Persistence as PersistenceModule
from nio.modules.scheduler import Job


class Persistence(object):

    """ A block mixin that provides persistence for some variables.

    To use, override the persisted_values function and define which variables
    on your class you wish to have persisted. The values should be strings
    that correspond to the variable names to be saved.

    """

    backup_interval = TimeDeltaProperty(
        visible=False, title='Backup Interval', default={"seconds": 60 * 60})
    load_from_persistence = BoolProperty(
        title='Load from Persistence?', default=True)

    def __init__(self):
        super().__init__()
        self._persistence = None
        self._backup_job = None

    def persisted_values(self):
        """ Return a list containing the values to be persisted.

        This function should be overriden in a Block that wishes to use
        persistence. Return a list of the names of the instance level variables
        to save.

        For example, if your block class has an instance level attribute called
        `_values` and you wish to save it, you would return this list:
            ["_values"]
        """
        return []

    def _load(self):
        """ Load the values from persistence """
        self.logger.debug("Loading from persistence")
        for persisted_var in self.persisted_values():
            if self._persistence.has_key(persisted_var):
                loaded = self._persistence.load(persisted_var)
                self.logger.debug("Loaded value {} for attribute {}".format(
                    loaded, persisted_var))
                # Set the loaded value to the attribute on this class
                setattr(self, persisted_var, loaded)

    def _save(self):
        """ Save the values to persistence """
        self.logger.debug("Saving to persistence")
        for persisted_var in self.persisted_values():
            self._persistence.store(
                persisted_var, getattr(self, persisted_var))
        self._persistence.save()

    def configure(self, context):
        super().configure(context)
        # Create a persistence object using the block's name
        self._persistence = PersistenceModule(self.name())
        if self.load_from_persistence():
            self._load()

    def start(self):
        super().start()

        # If we have configured a backup interval, set up the job to
        # periodically save to persistence
        if self.backup_interval().total_seconds() > 0:
            self._backup_job = Job(self._save, self.backup_interval(), True)

    def stop(self):
        if self._backup_job:
            self._backup_job.cancel()

        # Do one last save before stopping
        self._save()
        super().stop()



from nio.util.logging import get_nio_logger
from nio.util.scheduler.scheduler_helper import SchedulerHelper
from nio.util.scheduler.scheduler_thread import SchedulerThread


class Scheduler(object):

    """Singleton class to maintain and interact with an underlying Scheduler"""

    _scheduler_thread = None
    logger = None

    _sched_min_delta = 0.1
    _sched_resolution = 0.1

    @classmethod
    def _get_logger(cls):
        if cls.logger is None:
            cls.logger = get_nio_logger("NIOScheduler")
        return cls.logger

    @classmethod
    def configure(cls, context):
        # Load in the minimum delta and resolution from the config
        cls._sched_min_delta = context.min_interval
        cls._sched_resolution = context.resolution

    @classmethod
    def schedule_task(cls, target, delta, repeatable, *args, **kwargs):
        """ Add the given task to the Scheduler.

        Args:
            target (callable): The task to be scheduled.
            delta (timedelta): The scheduling interval. The scheduler will
                wait for this interval before running 'target' for the first
                time.
            repeatable (bool): When False, 'target' is run only once.
                Otherwise, it is run repeatedly at an interval defined by
                'delta'.
            args: Positional arguments to be passed to 'target'.
            kwargs: Keyword arguments to be passed to 'target'.

        Returns:
            job (Job): The Job object.

        """
        if cls._scheduler_thread is None:
            cls.start()

        return cls._scheduler_thread.scheduler.add(target, delta, repeatable,
                                                   *args, **kwargs)

    @classmethod
    def unschedule(cls, job):
        """Remove a job from the scheduler.

        If the given job is not currently scheduled, this method has no effect.

        Args:
            job (APScheduler Job): The job to remove.

        Returns:
            None

        """
        cls._get_logger().debug("Un-scheduling %s" % job)
        cls._scheduler_thread.scheduler.cancel(job)

    @classmethod
    def shutdown(cls):
        if cls._scheduler_thread:
            cls._scheduler_thread.stop()
        cls._scheduler_thread = None

    @classmethod
    def start(cls):
        try:
            if cls._scheduler_thread is None:
                cls._get_logger().info("Starting custom Scheduler")
                cls._scheduler_thread = SchedulerThread(SchedulerHelper(
                    resolution=cls._sched_resolution,
                    min_delta=cls._sched_min_delta))
                cls._scheduler_thread.start()
        except Exception:  # pragma: no cover (exception from ap)
            cls._get_logger().exception("Scheduler failed to start")



from nio.modules.proxy import ModuleProxy


class Unauthorized(Exception):

    """ Unauthorized security Exception

    Exception triggered when a user is not authorized to perform a task
    """
    pass


class Authorizer(ModuleProxy):

    @classmethod
    def authorize(cls, user, task):
        """ Ensure a user can perform a task.

        Args:
            user (User): A user object, likely returned by the Authenticator
            task (SecureTask): A task this user can or cannot perform

        Returns:
            None: The method will pass successfully if the user is authorized

        Raises:
            Unauthorized: if the user cannot perform the specified task
        """
        raise NotImplementedError()

    @classmethod
    def authorize_multiple(cls, user, *tasks, meet_all=True):
        """ Authorize a user against multiple tasks.

        This method is provided as a helper to authorize a user against
        multiple tasks. It will call Authorizer.authorize. You may override
        this in your implementation if it is needed for some reason, but it
        is recommended to leave it.

        Example:
            This example ensures that the user can perform t1 OR t2.
            >>> Authorizer.authorize_multiple(
                    User('test'), SecureTask('t1'), SecureTask('t2'),
                    meet_all=False)

        Args:
            user (User): A user object to check permissions against
            *tasks (list of SecureTasks): In a *args strucutre, pass in any
                number of SecureTask objects to check that the user can perform
            meet_all (bool): Whether or not the user must be able to perform
                all of the tasks or just one. The method will short-circuit
                as expected for either value of this bool.

        Returns:
            None: Same as the Authorizer.authorize method

        Raises:
            Unauthorized: Just like Authorizer.authorize, this will raise
                an exception if the user cannot perform the tasks.
        """
        for task in tasks:
            if meet_all:
                # Since they all need to pass, just try them, if one fails
                # it will raise the exception here
                cls.authorize(user, task)
            else:
                try:
                    cls.authorize(user, task)
                    # We need any of them to pass and this one has, let's
                    # break out and call it done
                    break
                except Unauthorized:
                    # This one didn't work, maybe the next one will, gotta
                    # keep trying
                    pass
        else:
            # Exhausting all of the options is good if we want to meet all.
            # However, if we want to meet any (meet_all = False), then that
            # means that the loop never broke out and we should raise our
            # exception
            if not meet_all:
                raise Unauthorized()



""" Discoverable decorator

Classes marked as Discoverable allow the system to identify them and register
them. While any class can technically be marked discoverable, n.io will tend
to only care about blocks, services, or components that are discoverable.

Note: A class being marked discoverable will not mark its subclasses as
discoverable. Each subclass must be explicitly marked with the discoverable
decorator.

To mark a class as discoverable, use the parameter-less decorator
`discoverable`::

    from nio import discoverable, Block

    @discoverable
    class MyBlock(Block):
        pass

"""


def discoverable(_class):
    """The decorator method to be called on the class object.

    This method will set the proper discoverable type to the class. It
    should return the class passed in, according to the decorator spec.
    """

    # Set the attribute to the class name, to prevent subclasses from also
    # being discoverable.
    setattr(_class, _get_discoverable_attribute(_class), True)
    return _class


def is_class_discoverable(_class):
    """ Returns true if the class is marked discoverable

    Note: discoverability does not extend to subclasses. Each subclass must
    explicitly be marked as discoverable

    Args:
        _class (class): The class in question, should be a class object

    Returns:
        bool: True if the class is discoverable
    """
    return bool(getattr(_class, _get_discoverable_attribute(_class), False))


def _get_discoverable_attribute(_class):
    """ Get an attribute to set on a class to consider it discoverable """
    return "__{}_is_discoverable".format(_class.__name__)



from nio.modules.proxy import ModuleProxy


class WebEngine(ModuleProxy):

    @classmethod
    def get(cls, port, host, configuration=None):
        """ Instantiates a web server

        Args:
            port (int): Port listening for requests
            host (str): Web server host address.
            configuration (dict): Contains additional server configuration,
                might contain additional ssl settings such as:
                ssl_certificate, ssl_private_key, ssl_certificate_chain

        Returns:
            The server if it already exists, or a new instantiated server

        Raises:
            TypeError if there port is None

        """
        raise NotImplementedError()

    @classmethod
    def remove_server(cls, server):
        """ Stops and remove a server

        Args:
            server (WebServer): Server to remove

        """
        raise NotImplementedError()

    @classmethod
    def get_servers(cls):
        """ Provides a list of running web servers """
        raise NotImplementedError()

    @classmethod
    def configure(cls, configuration):
        """ Configure Web Engine

        Args:
            configuration: Engine configuration
        """
        raise NotImplementedError()

    @classmethod
    def get_name(cls):
        """ Provides engine name """
        raise NotImplementedError()

    @classmethod
    def get_version(cls):
        """ Provides engine version """
        raise NotImplementedError()

    @classmethod
    def start(cls, callback=None):
        """ Starts Web engine

        Starts a web engine based on previously set configuration

        Args:
            callback: optional engine entry point

        """
        raise NotImplementedError()

    @classmethod
    def stop(cls):
        """ Stops engine and running Web Servers """
        raise NotImplementedError()

    @classmethod
    def block(cls):
        """ Blocks the Web engine

        This method is normally called from the main thread to block
        execution since likely each request will be handled by a separate
        thread.

        When stopping the engine, it is expected for it wait for all currently
        executing requests/threads to terminate.

        """
        raise NotImplementedError()



from threading import RLock

from nio.modules.communication.publisher import Publisher


class PublisherProxy(object):

    _publisher = None
    _publisher_lock = RLock()
    # establish that publisher is ready for use
    _publisher_ready = False
    # many loggers get initialized however, only once the proxy is initialized
    _initialized = False
    # publishing criteria
    _topics = {}

    @classmethod
    def init(cls, topics):
        """ Initializes the proxy

        Args:
            topics: topics to use when publishing signals

        """
        if not cls._initialized:
            # proxy initialization happens only once
            cls._topics = topics
            cls._initialized = True

    @classmethod
    def publish(cls, signals):
        """ Publishes logging signals.

        This method ensures that publisher is ready to send signals, it
        might happen that a call to publish 'logging' signals might come before
        publisher is ready since logging is initialized before communication


        Args:
            signals: signals to publish

        """
        cls._ensure_publisher_ready()
        # publish when all conditions are met
        if cls._publisher_ready:
            with cls._publisher_lock:
                cls._publisher.send(signals)

    @classmethod
    def close(cls):
        """ Closes publisher by releasing all resources """

        if cls._publisher:
            try:
                cls._publisher.close()
            except:
                pass
            finally:
                cls._publisher = None
                cls._publisher_ready = False
        cls._initialized = False

    @classmethod
    def _ensure_publisher_ready(cls):
        """ Ensures publisher is ready by attempting to create Publisher,
        which will happen successfully once communication module is initialized
        """
        if not cls._publisher_ready:
            try:
                cls._publisher = Publisher(**cls._topics)
                cls._publisher.open()
                cls._publisher_ready = True
            except:
                return



""" An object that keeps track of modules and configurations.

A ModuleInitializer will have a list of module (nio.modules.module.Module)
objects as well as what configuration contexts they should be initialized
with. When instructed to initialize, the ModuleInitializer will determine
the proper order to initialize the modules based on the module class
definitions.
"""
from nio.util.logging import get_nio_logger
from nio.modules.proxy import ProxyNotProxied, ProxyAlreadyProxied
from nio.modules.module import Module
from nio.modules.context import ModuleContext


class ModuleInitializer(object):

    """ A class that proxies and unproxies modules """

    def __init__(self):
        super().__init__()
        self.logger = get_nio_logger('ModuleInitializer')
        self._registered_modules = []
        self._initialized_modules = []

    def initialize(self, safe=False):
        """ Initialize and proxy any registered proxy implmentations.

        This function will proxy any implmentations that have been registered
        with this ModuleInitializer.

        Args:
            safe (bool): True if you want the initialize to unproxy the proxy
                first, before trying to proxy it. Defaults to False
        """
        for module, context in sorted(
                self._registered_modules,
                key=lambda mod: mod[0].get_module_order()):
            if safe:
                try:
                    module.initialize(context)
                except ProxyAlreadyProxied:
                    self.logger.warning(
                        "Module {} was already proxied, unproxying and trying "
                        "again".format(module.__class__.__name__))
                    module.finalize()
                    module.initialize(context)
            else:
                module.initialize(context)
            # Add this module to the list of modules that have been proxied
            self._initialized_modules.append(module)

    def register_module(self, module, context):
        """ Register an implmentation of the WebEngine proxy

        The class registered here will be proxied when the module initializer
        is initialized. The implementation class can also be passed to that
        method directly.

        Args:
            implmentation_class (class): A class containing the implementation
                for the WebEngine proxy

        Raises:
            TypeError: if the implementation class is not a class
        """
        if not isinstance(module, Module):
            raise TypeError("Registered module must be an instance of Module")
        if not isinstance(context, ModuleContext):
            raise TypeError("Context must be an instance of ModuleContext")
        self._registered_modules.append((module, context))

    def finalize(self, safe=False):
        """ De-register and unproxy the registered modules on this initializer

        This will finalize the modules in the reverse order that they
        were proxied.

        Args:
            safe (bool): True if you want the finalize to succeed even if
                the modules are already unproxied. Defaults to False
        """
        for module in reversed(self._initialized_modules):
            if safe:
                try:
                    module.finalize()
                except ProxyNotProxied:
                    self.logger.warning(
                        "Ignoring ProxyNotProxied while finalizing {}".format(
                            module.__class__.__name__))
            else:
                module.finalize()
        self._initialized_modules[:] = []



import re
from setuptools import setup, find_packages

with open('nio/__init__.py', 'r') as fd:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        fd.read(), re.MULTILINE).group(1)

if not version:
    raise RuntimeError('Cannot find version information')

setup(
    name='nio',
    description='A framework for building blocks for n.io',
    keywords=['nio', 'n.io'],
    version=version,
    packages=find_packages(
        exclude=['*.tests', '*.tests.*', 'tests.*', 'tests']),

    # No dependenices, believe it!
    install_requires=[],

    tests_require=[
        'requests>=2.3.0'
    ],

    author='n.io',
    author_email='info@n.io',
    url='http://n.io'
)



from threading import Thread


class SchedulerThread(Thread):
    """ host the scheduler.
    """

    def __init__(self, scheduler):
        """ Initializes the thread.

        Args:
            scheduler (Scheduler): scheduler instance it will run
        """
        super(SchedulerThread, self).__init__(daemon=True)

        self._scheduler = scheduler

    def run(self):
        """ Scheduler thread entry point.

        Starts the scheduler

        """
        self._scheduler.start()

    def stop(self):
        """ Stops the thread and the scheduler
        """
        self._scheduler.stop()

    @property
    def scheduler(self):
        return self._scheduler



from time import sleep
from nio.util.threading.thread import NIOThread


def spawn(target, *args, **kwargs):
    """ Executes given target in a new thread

    Args:
        target: method or function to execute
        *args: positional arguments in 'target'
        **kwargs: keyword arguments in 'target'

    Returns:
        newly created thread

    """
    args = (target,) + args
    t = NIOThread(target=_spawn_target, args=args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t


def _spawn_target(target, *args, **kwargs):
    SLEEP_TIME_BEFORE_LAUNCHING_THREAD = 0.00001
    sleep(SLEEP_TIME_BEFORE_LAUNCHING_THREAD)
    return target(*args, **kwargs)
